/*
 -------------------------------------------------------------------------
 Copyright (c) 2002-2006, XIMETA, Inc., FREMONT, CA, USA.
 All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary 
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright 
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products 
      built using this software without specific written permission. 
      
 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.
 
 DISCLAIMER

 This software is provided 'as is' with no explcit or implied warranties
 in respect of any properties, including, but not limited to, correctness 
 and fitness for purpose.
 -------------------------------------------------------------------------
*/
#include <linux_ver.h>
#if LINUX_VERSION_25_ABOVE
#include <linux/module.h> // THIS_MODULE
#include <ndasuser/ndasuser.h>
#include <ndasuser/io.h>
#include <sal/debug.h>
#include <ndasdev.h>
#include <procfs.h>
#include "block.h"
#include "ops.h"

struct block_device_operations ndas_fops = {
    .owner =            THIS_MODULE, 
    .open =            ndop_open, 
    .release =        ndop_release, 
    .ioctl =            ndop_ioctl, 
    .media_changed = ndop_media_changed,
    .revalidate_disk = ndop_revalidate_disk
};

KMEM_CACHE* nblk_kmem_cache;
extern int ndas_io_unit;

static struct nbio* nbio_alloc_splited(struct request *req)
{
    char *data;
    struct bio *bio;
    struct bio_vec* bvec;
    int slot = SLOT_R(req);
    int i, len = 0, chunk_size = 1 << 9;
    struct nbio *lio = kmem_cache_alloc(nblk_kmem_cache, GFP_ATOMIC);
            /* size: NBIO_SIZE(req->nr_sectors */
    if ( !lio ) return NULL;

    lio->req = req;
    lio->nr_blocks = 0;
    lio->dev = NDAS_GET_SLOT_DEV(slot);
    rq_for_each_bio(bio, req) 
    {
        bio_for_each_segment(bvec, bio, i) 
        {
            int j = bvec->bv_len; 
            data = page_address(bvec->bv_page) + bvec->bv_offset;
            while ( j > 0 ) 
            {
                int idx = SPLIT_IDX_SLOT(lio->nr_blocks, req->nr_sectors, slot); 
                lio->blocks[idx].len = chunk_size;
                lio->blocks[idx].ptr = data;
                j -= chunk_size;
                data += chunk_size;
                lio->nr_blocks++;
            }
            len += bvec->bv_len;
        }
    }
    lio->nr_sectors = len >> 9;
    sal_assert( lio->nr_sectors == req->nr_sectors );
    return lio;
}
static inline 
struct nbio* nbio_alloc(struct request *req) 
{
    int i, len = 0;
    struct bio *bio;
    struct bio_vec* bvec;
    int slot = SLOT_R(req);
    struct nbio *lio = kmem_cache_alloc(nblk_kmem_cache, GFP_ATOMIC);
            /* size: NBIO_SIZE(req->nr_phys_segments) */
    if ( !lio ) return NULL;
    
    lio->req = req;
    lio->nr_blocks = 0;
    lio->dev = NDAS_GET_SLOT_DEV(slot);
    
    rq_for_each_bio(bio, req) 
    {
        bio_for_each_segment(bvec, bio, i) 
        {
#ifdef CONFIG_HIGHMEM

            if ( is_highmem(page_zone(bvec->bv_page)) ) {
                lio->blocks[lio->nr_blocks].private = bvec->bv_page;
                lio->blocks[lio->nr_blocks].ptr = kmap(bvec->bv_page) + bvec->bv_offset;
            } else {
                lio->blocks[lio->nr_blocks].private = NULL;
                lio->blocks[lio->nr_blocks].ptr = page_address(bvec->bv_page) + bvec->bv_offset;
            }
            lio->blocks[lio->nr_blocks].len = bvec->bv_len;
            lio->nr_blocks++;
            
#else
	    char *data;
            data = page_address(bvec->bv_page) + bvec->bv_offset;
            if ( NBIO_ADJACENT(lio,data) )
                lio->blocks[lio->nr_blocks - 1].len += bvec->bv_len;
            else {
                lio->blocks[lio->nr_blocks].len = bvec->bv_len;
                lio->blocks[lio->nr_blocks].ptr = data;
                lio->nr_blocks++;
            }
#endif
            len += bvec->bv_len;
        }
        
    }
    lio->nr_sectors = len >> 9;
    //unlikely( lio->nr_blocks <= req->nr_phys_segments);
    sal_assert( lio->nr_sectors == req->nr_sectors);
    return lio;
}

static int _is_not_recoverable(ndas_error_t err) {
    return (err != NDAS_ERROR_INVALID_RANGE_REQUEST &&
        err != NDAS_ERROR_READONLY &&
        err != NDAS_ERROR_READONLY &&
        err != NDAS_ERROR_OUT_OF_MEMORY &&
        err != NDAS_ERROR_HARDWARE_DEFECT &&
        err != NDAS_ERROR_BAD_SECTOR);
}

static void _end_io(ndas_error_t err, struct request *req, struct ndas_slot *dev, size_t nr_sectors) 
{
    unsigned long flags;
    struct request_queue *q = req->q;
    int uptodate = NDAS_SUCCESS(err) ? 1 : 0;
     
    spin_lock_irqsave(q->queue_lock, flags);
    if ( !uptodate ) {
        printk("ndas: IO error occurred at sector %d(slot %d): %s\n", 
            (int)req->sector,
            SLOT_R(req),
            NDAS_GET_STRING_ERROR(err));
        req->errors++;
        if ( _is_not_recoverable(err) ) {
            if (!test_bit(NDAS_FLAG_QUEUE_SUSPENDED, &dev->queue_flags)) {
                set_bit(NDAS_FLAG_QUEUE_SUSPENDED, &dev->queue_flags);
            }
        }
    }

    if (!end_that_request_first(req, uptodate, nr_sectors)) {
        END_THAT_REQUEST_LAST(req, uptodate);
    } 

    spin_unlock_irqrestore(q->queue_lock, flags);
}
/* signalled when the NDAS io is completed 
 */
NDAS_CALL
void nblk_end_io(int slot, ndas_error_t err, struct nbio* lio)
{
    struct request *req = lio->req;
#ifdef CONFIG_HIGHMEM
    int i;
    for (i = 0; i < lio->nr_blocks; i++) {
        if ( lio->blocks[i].private ) 
            kunmap((struct page *)lio->blocks[i].private);
    }
#endif     
    _end_io(err, req, lio->dev, lio->nr_sectors);
    kmem_cache_free(nblk_kmem_cache ,lio);
}
    
ndas_error_t nblk_handle_io(struct request *req) 
{
    ndas_error_t err;
    ndas_io_request request;
    int slot = SLOT_R(req);
    struct ndas_slot *sd = NDAS_GET_SLOT_DEV(slot);
    struct nbio *lio;

    if ( sd->info.io_splits == 1 )  {
        lio = nbio_alloc(req);
    } else {
        lio = nbio_alloc_splited(req);
    }
    if ( !lio ) {
        _end_io(NDAS_ERROR_OUT_OF_MEMORY, req, sd, req->nr_sectors);
        return NDAS_ERROR_OUT_OF_MEMORY;
    }

    if (!sd->enabled) {
        err =  NDAS_ERROR_NOT_CONNECTED;
        goto out;
    }

    request.uio = lio->blocks;
    request.buf = NULL;
    request.nr_uio = lio->nr_blocks;
    request.num_sec = lio->nr_sectors;
    request.start_sec = req->sector;
    request.done = (ndas_io_done) nblk_end_io;
    request.done_arg = lio;

    switch ( REQ_DIR(req) ) {
    case READ:
    case READA:
        err = ndas_read(slot,&request);
        break;
    case WRITE:
        err = ndas_write(slot,&request);
        break;
    default:
        err = NDAS_ERROR_INVALID_OPERATION;
        printk("ndas: operation %u is not supported.\n", REQ_DIR(req));
    }
out:    
    if ( !NDAS_SUCCESS(err) ) {
        nblk_end_io(slot, err, lio);
    }
    return err;
}

#define NBLK_LOCK(q) ((q)->queue_lock)

/* block request procedure. 
   Entered with q->queue_lock locked 
*/
void nblk_request_proc(struct request_queue *q)
{
    struct request *req;
    ndas_error_t err = NDAS_OK;

    while((req = NBLK_NEXT_REQUEST(q)) != NULL) 
    {
        if (test_bit(NDAS_FLAG_QUEUE_SUSPENDED, &(NDAS_GET_SLOT_DEV(SLOT_R(req))->queue_flags)))  {
            /* Queue is suspended */
            blkdev_dequeue_request(req);
            REQUEST_FLAGS(req) |= REQ_QUIET;
            REQUEST_FLAGS(req) |= REQ_FAILFAST; /* do not retry */
            REQUEST_FLAGS(req) |= REQ_FAILED;
            while (end_that_request_first(req, 0, req->nr_sectors))
                     ;
            END_THAT_REQUEST_LAST(req, 0);

            continue;
        }
        if (! BLK_CHECK_VALID_STATUS(req)) {
            printk ("ndas: skip non-CMD request\n");
            end_request(req, 0);

            continue;
        }
        if ( BLK_ATTEMPT_WRITE_RDONLY_DEVICE(req) ) {
            REQUEST_FLAGS(req) |= REQ_QUIET | REQ_FAILFAST | REQ_FAILED;
            end_request(req, 0);

            continue;
        }
        
        blkdev_dequeue_request(req);
        spin_unlock_irq(NBLK_LOCK(q));
        err = nblk_handle_io(req);
        spin_lock_irq(NBLK_LOCK(q));

        if ( !NDAS_SUCCESS(err) ) {
            break;
		}
    }
}

int blk_init(void) 
{    
    int ret;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23))
    nblk_kmem_cache = kmem_cache_create("ndas_bio", MAX_NBIO_SIZE, 0, SLAB_HWCACHE_ALIGN,
        NULL);
#else
    nblk_kmem_cache = kmem_cache_create("ndas_bio", MAX_NBIO_SIZE, 0, SLAB_HWCACHE_ALIGN,
        NULL, NULL);
#endif
    if (!nblk_kmem_cache) {
        printk(KERN_ERR "ndas: unable to create kmem cache\n");
        ret = -ENOMEM;
        goto errout;
    }

    if ( (ret = register_blkdev(NDAS_BLK_MAJOR, "ndas")) < 0 ) {
        printk(KERN_ERR "ndas: unable to get major %d for ndas block device\n", NDAS_BLK_MAJOR);
        goto errout;
    }

#ifdef NDAS_HOTPLUG
    /* register NDAS bus */
    ret = bus_register(&ndas_bus_type);
    if ( ret ) {        
        printk("ndas: Unable to register bus ndas\n");    
        goto errout;
    }
    ndas_bus_type.devices.hotplug_ops = &nd_hotplug_ops;
    /* register NDAS driver */
    ret = driver_register(&v_ndas_driver.driver);
    if ( ret ) {
        printk("ndas: Unable to register driver ndas\n");
        goto errout;
    }
    driver_create_file(&v_ndas_driver.driver,&driver_attr_version);
#endif    
    
    printk("ndas: registered ndas device at major number %d\n", NDAS_BLK_MAJOR);
    return 0;
errout:

    unregister_blkdev(NDAS_BLK_MAJOR, "ndas");
    
    if (nblk_kmem_cache) {
        kmem_cache_destroy(nblk_kmem_cache);
        nblk_kmem_cache = NULL;
    }
    return ret;
}

/*
 * block device clean up
 */
int blk_cleanup(void)
{
    unregister_blkdev(NDAS_BLK_MAJOR, "ndas");
    if (nblk_kmem_cache) {
        kmem_cache_destroy(nblk_kmem_cache);
        nblk_kmem_cache = NULL;
    }
    return 0;
}


#ifdef NDAS_HOTPLUG
/* ndas device */

struct ndas_driver {
    struct device_driver driver;
};
/*
 * Set up "version" as a driver attribute.
 */
static ssize_t nddriver_show_version(struct device_driver *drv, char *buf)
{
    strcpy(buf, "1.1");
    return strlen("1.1") + 1;
}

static DRIVER_ATTR(version, S_IRUGO, nddriver_show_version, NULL);

static inline struct ndas_slot* to_ndas_dev(struct device* dev)
{
    return container_of(dev,struct ndas_slot, dev);
}

static int ndas_bus_match(struct device * dev, struct device_driver * drv)
{
    return 0;
}
static struct device * ndas_bus_add(struct device * parent, char * bus_id) {
    return NULL;
    
}
static int ndas_bus_hotplug (struct device *dev, char **envp, 
                int num_envp, char *buffer, int buffer_size)
{
    return 0;
}
static int ndas_bus_suspend(struct device * dev, u32 state)
{
    return 0;

}
static int ndas_bus_resume(struct device * dev)
{
    return 0;

}

static int nd_probe       (struct device * dev) {
    return 0;
}
static int nd_remove      (struct device * dev){
    return 0;
}
static void nd_stop     (struct device * dev){
}
static int nd_suspend     (struct device * dev, u32 state, u32 level){
    return 0;
}
static int nd_resume      (struct device * dev, u32 level){
    return 0;
}
struct bus_type ndas_bus_type = {
    .name = "ndas",
    .match = ndas_bus_match,
    .add = ndas_bus_add,    
    .hotplug = ndas_bus_hotplug,    
    .suspend = ndas_bus_suspend,
    .resume = ndas_bus_resume,
};
EXPORT_SYMBOL(ndas_bus_type);
static struct ndas_driver v_ndas_driver = {
    .driver = {
        .name = "ndas",
        .bus = &ndas_bus_type,
        .probe = nd_probe,
        .remove = nd_remove,
        .shutdown = nd_stop,
        .suspend = nd_suspend,
        .resume = nd_resume
    },
};

int nddev_filter(struct kset *kset, struct kobject *kobj)
{
    struct kobj_type *ktype = get_ktype(kobj);
    if (ktype == ndas_bus_type.devices.ktype) {
        struct device *dev = container_of(kobj, struct device,kobj); 
        if (dev->bus)
            return 1;
    }
    return 0;
}
char* nddev_name(struct kset *kset, struct kobject *kobj)
{
    struct device *dev = container_of(kobj, struct device,kobj);
    return dev->bus->name; 
}
int nddev_hotplug(struct kset *kset, struct kobject *kobj, char **envp,
        int num_envp, char *buffer, int buffer_size)
{
    //struct usb_interface *intf;
    struct ndas_slot *nd_adapter_dev;
    char *scratch;
    int i = 0;
    int length = 0;
    struct device *dev = container_of(kobj, struct device,kobj);
    
    /* ... */
    nd_adapter_dev = container_of(dev,struct ndas_slot, dev);
//    intf = to_usb_interface(dev);
//    usb_dev = interface_to_usbdev(intf);

    /* ... */
    scratch = buffer;
    envp[i++] = scratch;
    length += snprintf(scratch, buffer_size - length, "PRODUCT=%x",
               nd_adapter_dev->slot);
    if ((buffer_size - length <= 0) || (i >= num_envp))
        return -ENOMEM;
    ++length;
    scratch += length;

    /* ... */
    envp[i++] = NULL;
    return 0;
}

static struct kset_hotplug_ops nd_hotplug_ops = {
    .filter = nddev_filter,
    .name = nddev_name,
    .hotplug = nddev_hotplug,
};
#endif
#endif // 2.6

